/** \file cbp.h
 * \brief Implements CBP (Clamped BP)
 *
 * This algorithm uses configurable heuristics to choose a variable
 * x_i and a state x_i*. Inference is done with x_i "clamped" to x_i*
 * (e.g. conditional on x_i==x_i*), and also with the negation of this
 * condition. Clamping is done recursively up to a fixed number of
 * levels (other stopping criteria are also implemented, see
 * "recursion" property). The resulting approximate marginals are
 * combined using logZ estimates.
 *
 * See Eaton & Ghahramani, "Choosing a Variable to Clamp", AISTATS 2009
 */
/* 
Copyright 2009 Frederik Eaton
*/

#ifndef ____defined_libdai_cbp_h__
#define ____defined_libdai_cbp_h__

#include <dai/daialg.h>

#include <dai/cbp.h>
#include <dai/bbp.h>
#include <dai/clamp_util.h>

namespace dai {

using namespace std;

/// Find a variable to clamp using BBP (goes with maximum adjoint)
/// \see BBP
pair<size_t, size_t> bbpFindClampVar(const InfAlg &in_bp, bool clampingVar,
            const PropertySet &bbp_props, bbp_cfn_t cfn, 
            Real *maxVarOut);

/// Class for CBP ("clamped belief propagation")
class CBP : public DAIAlgFG {
    vector<Factor> _beliefsV;
    vector<Factor> _beliefsF;
    double _logZ;
    double _est_logZ;
    double _old_est_logZ;

    double _sum_level;
    size_t _num_leaves;

    bbp_cfn_t BBP_cost_function() {
      return props.bbp_cfn;
    }

    void printDebugInfo();

    /// Called by 'run', and by itself. Implements the main algorithm. 
    /// Chooses a variable to clamp, recurses, combines the logZ and
    /// beliefs estimates of the children, and returns the improved
    /// estimates in \a lz_out and \a beliefs_out to its parent
    void runRecurse(InfAlg *bp,
                    double orig_logZ,
                    vector<size_t> clamped_vars_list,
                    size_t &num_leaves,
                    double &sum_level,
                    Real &lz_out,
                    vector<Factor>& beliefs_out);

    /// Choose the next variable to clamp, given a converged InfAlg (\a bp),
    /// and a vector of variables that are already clamped (\a
    /// clamped_vars_list). Returns the chosen variable in \a i, and
    /// the set of states in \a xis. If \a maxVarOut is non-NULL and
    /// props.choose==CHOOSE_BBP then it is used to store the
    /// adjoint of the chosen variable
    virtual bool chooseNextClampVar(InfAlg* bp,
                                    vector<size_t> &clamped_vars_list,
                                    size_t &i, vector<size_t> &xis, Real *maxVarOut);

    /// \brief Return the InfAlg to use at each step of the recursion. 
    /// \todo At present, only returns a BP instance
    InfAlg* getInfAlg();

    size_t _iters;
    double _maxdiff;

    void setBeliefs(const vector<Factor>& bs, double logZ) {
        size_t i=0;
        _beliefsV.clear(); _beliefsV.reserve(nrVars());
        _beliefsF.clear(); _beliefsF.reserve(nrFactors());
        for(i=0; i<nrVars(); i++) _beliefsV.push_back(bs[i]);
        for(; i<nrVars()+nrFactors(); i++) _beliefsF.push_back(bs[i]);
        _logZ = logZ;
    }

    void construct();

  public:

    //----------------------------------------------------------------

    // construct CBP object from FactorGraph
    CBP(const FactorGraph &fg, const PropertySet &opts) : DAIAlgFG(fg) {
        props.set(opts);
        construct();
    }

    static const char *Name;

    /// @name General InfAlg interface
    //@{
    virtual CBP* clone() const { return new CBP(*this); }
    virtual CBP* create() const { DAI_THROW(NOT_IMPLEMENTED); }
    virtual std::string identify() const { return string(Name) + props.toString(); }
    virtual Factor belief (const Var &n) const { return _beliefsV[findVar(n)]; }
    virtual Factor belief (const VarSet &) const { assert(0); }
    virtual vector<Factor> beliefs() const { return concat(_beliefsV, _beliefsF); }
    virtual Real logZ() const { return _logZ; }
    virtual void init() {};
    virtual void init( const VarSet & ) {};
    virtual double run();
    virtual double maxDiff() const { return _maxdiff; }
    virtual size_t Iterations() const { return _iters; }
    //@}

    Factor beliefV (size_t i) const { return _beliefsV[i]; }
    Factor beliefF (size_t I) const { return _beliefsF[I]; }

    //----------------------------------------------------------------

 public:
/* PROPERTIES(props,CBP) {
        typedef BP::Properties::UpdateType UpdateType;
        DAI_ENUM(RecurseType,REC_FIXED,REC_LOGZ,REC_BDIFF);
        DAI_ENUM(ChooseMethodType,CHOOSE_RANDOM,CHOOSE_MAXENT,CHOOSE_BBP,CHOOSE_BP_L1,CHOOSE_BP_CFN);
        DAI_ENUM(ClampType,CLAMP_VAR,CLAMP_FACTOR);
        
        size_t verbose;

        /// Tolerance to use in BP
        double tol;
        /// Update style for BP
        UpdateType updates;
        /// Maximum number of iterations for BP
        size_t maxiter;

        /// Tolerance to use for controlling recursion depth (\a recurse
        /// is REC_LOGZ or REC_BDIFF)
        double rec_tol;
        /// Maximum number of levels of recursion (\a recurse is REC_FIXED)
        size_t max_levels;
        /// If choose=CHOOSE_BBP and maximum adjoint is less than this value, don't recurse
        double min_max_adj;
        /// Heuristic for choosing clamping variable
        ChooseMethodType choose;
        /// Method for deciding when to stop recursing
        RecurseType recursion;
        /// Whether to clamp variables or factors
        ClampType clamp;
        /// Properties to pass to BBP
        PropertySet bbp_props;
        /// Cost function to use for BBP
        bbp_cfn_t bbp_cfn;
        /// Random seed
        size_t rand_seed;
   }
*/
/* {{{ GENERATED CODE: DO NOT EDIT. Created by 
    ./scripts/regenerate-properties include/dai/cbp.h src/cbp.cpp 
*/
    struct Properties {
        typedef BP::Properties::UpdateType UpdateType;
        DAI_ENUM(RecurseType,REC_FIXED,REC_LOGZ,REC_BDIFF);
        DAI_ENUM(ChooseMethodType,CHOOSE_RANDOM,CHOOSE_MAXENT,CHOOSE_BBP,CHOOSE_BP_L1,CHOOSE_BP_CFN);
        DAI_ENUM(ClampType,CLAMP_VAR,CLAMP_FACTOR);
        size_t verbose;
        /// Tolerance to use in BP
        double tol;
        /// Update style for BP
        UpdateType updates;
        /// Maximum number of iterations for BP
        size_t maxiter;
        /// Tolerance to use for controlling recursion depth (\a recurse
/// is REC_LOGZ or REC_BDIFF)
        double rec_tol;
        /// Maximum number of levels of recursion (\a recurse is REC_FIXED)
        size_t max_levels;
        /// If choose=CHOOSE_BBP and maximum adjoint is less than this value, don't recurse
        double min_max_adj;
        /// Heuristic for choosing clamping variable
        ChooseMethodType choose;
        /// Method for deciding when to stop recursing
        RecurseType recursion;
        /// Whether to clamp variables or factors
        ClampType clamp;
        /// Properties to pass to BBP
        PropertySet bbp_props;
        /// Cost function to use for BBP
        bbp_cfn_t bbp_cfn;
        /// Random seed
        size_t rand_seed;

        /// Set members from PropertySet
        void set(const PropertySet &opts);
        /// Get members into PropertySet
        PropertySet get() const;
        /// Convert to a string which can be parsed as a PropertySet
        string toString() const;
    } props;
/* }}} END OF GENERATED CODE */

    Properties::ChooseMethodType ChooseMethod() { return props.choose; }
    Properties::RecurseType Recursion() { return props.recursion; }
    Properties::ClampType Clamping() { return props.clamp; }
    size_t maxClampLevel() { return props.max_levels; }
    double minMaxAdj() { return props.min_max_adj; }
    double recTol() { return props.rec_tol; }
};

}

#endif
